@@ -950,6 +950,9 @@ static int get_metadata_lazy_checked(
 			dis_free(*metadata);
 	}
 
+	if(current >= 3)
+		return FALSE;
+
 	return TRUE;
 }
 
@@ -1032,6 +1035,9 @@ static int get_eow_check_valid(
 		}
 	}
 
+	if(current == 2)
+		return FALSE;
+
 	return TRUE;
 }
 
